"use strict"

class FTX extends Exchange {
	static meta = {
		aliases: [
			"FTX",
		],
		endpoint: "https://ftx.com",
		fields: {
			public: {
				label: "API Key",
			},
			private: {
				label: "API Secret",
			},
			sub: {
				label: "Subaccount",
				message: "Optional – FTX subaccount to be addressed through this PV account (case sensitive!)",
				type: "text",
				optional: true
			}
		},
		name: "FTX",
		patterns: [],
		permissions: {
			origins: [
				"https://ftx.com/api/*",
				"https://otc.ftx.com/api/time"
			],
		},
		subscriptions: {
			active: [
				"e3fdd77de3a682bfcfd8dc3000a25a5b",
				"4db6261b067dabf8ffb0924c18d95d06",
			],
			inactive: [
				"ftxxx1",
				"ftxxx2",
			],
		},
		website: "https://ftx.com/#a=1520859",
		recvWindow: 30 * 1000,
		desc: "FTX Cryptocurrency Derivatives Exchange, built by traders, for traders. Buy and sell BTC, ETH, USDT, BNB futures and index futures with up to 101x leverage.",
		defaultSymbol: "USD",
		testSymbol: "BTC-PERP",
		margin: true,
		spot: true
	}
	constructor() {
		super(FTX.meta)
	}

	*req(method, resource, params, sign = true) {
		resource = "/api" + resource
		if (method === 'GET' && params) {
			resource += '?' + serialize(params).replace(/%20/g, '+')
			params = null
		}
		params = params ? JSON.stringify(params) : ''

		const ts = this.getTimestamp()
		const payload = ts + method + resource + params
		const headers = {"Content-Type": "application/json"}
		if (sign) {
			const credentials = this.getCredentials("private")
			let sha = new jsSHA("SHA-256", "TEXT")
			sha.setHMACKey(credentials.private, "TEXT")
			sha.update(payload)

			headers["FTX-KEY"] = credentials.public
			headers["FTX-SIGN"] = sha.getHMAC("HEX")
			headers["FTX-TS"] = ts
			if (credentials.sub) {
				headers["FTX-SUBACCOUNT"] = credentials.sub
			}
		} else this.hasAccess()

		const response = yield request.bind(this, method, this.meta.endpoint + resource, params, headers)
		if (!response) {
			throw new Error("Unknown connection error with API endpoint!")
		}
		if (response.error || typeof response.result === 'undefined') {
			throw new Error(response.error || "Unknown data error with API endpoint!")
		}
		return response.result
	}

	*time() {
		const result = (yield request.bind(this, 'GET', "https://otc.ftx.com/api/time", null, null)).result
		return new Date(result).getTime()
	}

	*account(currency, isMargin) {
		let balances = {}
		if (!isMargin) {
			const wallet = yield* this.req('GET', "/wallet/balances")
			wallet.forEach && wallet.forEach(item => {
				const coin = item.coin.toUpperCase()
				if (Number(item.total) || coin == currency) balances[coin] = {
					available: Number(item.availableWithoutBorrow || item.free),
					balance: Number(item.total)
				}
			})
		} else {
			const account = yield* this.req('GET', "/account")
			balances[this.meta.defaultSymbol] = {
				available: Number(account.freeCollateral),
				balance: Number(account.collateral),
				leverage: Number(account.leverage)
			}
		}
		return balances
	}

	*symbolInfo(symbol) {
		const symbolCache = this.getSymbolCache()
		if (symbolCache._refresh) {
			const response = yield* this.req('GET', "/markets", null, false)

			response.forEach(info => {
				info.pricePrecision = decimals(info.priceIncrement) + (!info.priceIncrement.toString().endsWith('1') ? info.priceIncrement % 1 : 0)
				info.quantityPrecision = decimals(info.sizeIncrement) + (!info.sizeIncrement.toString().endsWith('1') ? info.sizeIncrement % 1 : 0)
				symbolCache[this.sym(info.name)] = info
			})
			this.updatedSymbolCache()
		}
		if (!symbol) return symbolCache

		const lookup = this.sym(symbol)
		if (!symbolCache[lookup]) {
			throw new Error(`Unknown market symbol: ${symbol} – Use the 'Symbols' button in the exchange settings or alert editor to browse for valid symbols!`)
		}
		return symbolCache[lookup]
	}

	*symbolTicker(symbol) {
		const ticker = yield* this.req('GET', "/markets/"+symbol, null, false)
		return ticker && {
			ask: Number(ticker.ask),
			bid: Number(ticker.bid),
			mid: (Number(ticker.ask) + Number(ticker.bid)) / 2,
			last: Number(ticker.last)
		}
	}

	checkLev(lev, c) {
		if (!lev) return lev
		const levs = [1, 3, 5, 10, 20, 50, 100, 101]
		const newLev = levs.reduce((prev, curr) => Math.abs(curr - lev) < Math.abs(prev - lev) && curr <= lev ? curr : prev)
		if (newLev != lev) {
			this.info(`rounded down to next valid leverage of ${newLev}x`)
			c && (c.f = lev / newLev)
		}
		return newLev
	}

	*trade(cmd, ohlc = {}) {
		const market = yield* this.symbolInfo(cmd.s)
		const ticker = (cmd.pc && !cmd.up && this.getPrice(market.name)) || (yield* this.symbolTicker(market.name))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.name} is not available!`)
		}
		this.checkSlip(cmd, ohlc, market.name, ticker)
		if (cmd.up) {
			this.updatePrice(market.name, ticker)
			if (!cmd.b && !cmd.ub) {
				return ticker
			}
		}

		const isSpot = market.type === 'spot'
		const currency = cmd.currency = (isSpot && cmd.isBid ? market.quoteCurrency : market.baseCurrency) || this.meta.defaultSymbol
		if (!isSpot) cmd.currency += marginSuffix
		let balance = {available: 0, balance: 0}
		if ((cmd.y !== "possize" && cmd.q.wasPercent()) || cmd.ub) {
			balance = cmd.bc && !cmd.ub && this.getBalance(currency, !isSpot)
			const balances = !balance && (yield* this.account(currency, !isSpot))
			balance = balance || balances[currency]
			this.checkBalance(cmd, balance, currency)
			if (!cmd.bc || cmd.ub) {
				this.updateBalance(balances, currency, !isSpot)
				if (cmd.ub) {
					return balance
				}
			}
		}

		const first = ticker[cmd.pr] || ohlc[cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.fp ? cmd.fp.resolve(market.pricePrecision) : cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(market.pricePrecision)
		if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated, use minp= to set minimum)")
		}

		ohlc[cmd.lr] && this.info(`using custom leverage multiplier: ${ohlc[cmd.lr]} (${cmd.lr})`)
		const c = {f: 1.0}
		const leverage = (cmd._lev = cmd.lc || (!isSpot && (this.checkLev(cmd.l * (ohlc[cmd.lr] || 1), c) || balance.leverage || 
			(this.getBalance(this.meta.defaultSymbol, !isSpot, true) || {}).leverage)) || 1)
		const available = cmd.y === "equity" ? balance.balance : balance.available
		const contracts = ohlc[cmd.y] || Math.abs(available) * leverage * c.f / (isSpot && cmd.isAsk ? 1 : price)

		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		let order = {
			market: market.name,
			side: cmd.isBid ? "buy" : "sell",
			price: cmd.t !== "market" ? price : null,
			type: cmd.t === "market" ? "market" : "limit",
			size: cmd.q._().mul(!cmd.q.wasPercent() ? (isSpot ? 1 : leverage * c.f) / (cmd.u === "currency" ? price : 1) : 1)
				.reference(contracts).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.quantityPrecision)
		}
		if (cmd.lq && ohlc.left) {
			this.info("using leftover quantity from last order: "+ohlc.left)
			order.size = ohlc.left.toFixed(market.quantityPrecision)
		}
		if (market.minProvideSize && order.size < market.minProvideSize) {
			this.warn(`order quantity below instrument minimum of ${market.minProvideSize} – use minq= to set minimum!`)
		} else if (order.size <= 0) {
			this.warn("your q parameter might be incorrect! (quantity of 0 calculated, use minq= to set minimum)")
		}
		if (cmd.r) order.reduceOnly = true
		if (cmd.t === "ioc") order.ioc = true
		if (cmd.t === "post") order.postOnly = true

		if (cmd.tp) {
			cmd.tpref && cmd.tpref !== "price" && this.info(`using custom TP reference: ${ohlc[cmd.tpref] || ticker[cmd.tpref] || "NOT FOUND!"} (${cmd.tpref})`)
			cmd.tpref !== "price" && (price = (cmd.tpref && (ohlc[cmd.tpref] || ticker[cmd.tpref])) || first)
			order.type = "takeProfit"
			order.triggerPrice = cmd.tp._().relative(price).resolve(market.pricePrecision)
			order.price && (order.orderPrice = order.price)
			delete order.price
			delete order.ioc
			delete order.postOnly
			order.retryUntilFilled = cmd.rf === false ? false : true
		} else if (cmd.sl || cmd.so) {
			const ref = cmd.slref || cmd.soref
			ref && ref !== "price" && this.info(`using custom Stop reference: ${ohlc[ref] || ticker[ref] || "NOT FOUND!"} (${ref})`)
			ref !== "price" && (price = (ref && (ohlc[ref] || ticker[ref])) || first)
			order.type = "stop"
			order.triggerPrice = (cmd.sl || cmd.so)._().relative(price).resolve(market.pricePrecision)
			order.price && (order.orderPrice = order.price)
			delete order.price
			delete order.ioc
			delete order.postOnly
			order.retryUntilFilled = cmd.rf === false ? false : true
		} else if (cmd.ts) {
			cmd.tsref && cmd.tsref !== "price" && this.info(`using custom TS reference: ${ohlc[cmd.tsref] || ticker[cmd.tsref] || "NOT FOUND!"} (${cmd.tsref})`)
			cmd.tsref !== "price" && (price = (cmd.tsref && (ohlc[cmd.tsref] || ticker[cmd.tsref])) || first)
			order.type = "trailingStop"
			order.trailValue = cmd.ts._().reference(price).abs().mul(cmd.isAsk ? -1 : 1).resolve(market.pricePrecision)
			delete order.price
			delete order.ioc
			delete order.postOnly
			order.retryUntilFilled = cmd.rf === false ? false : true
		} else if (cmd.id) {
			order.clientId = this.uniqueID(cmd.id)
		}

		this.info(`placing ${currency} order${isSpot?'':` @ ${leverage}x`}: `+stringify(order))
		if (cmd.d) {
			this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}

		try {
			if (!isSpot && cmd.l !== undefined) {
				this.info(`setting account leverage to ${leverage}x...`)
				yield* this.req('POST', "/account/leverage", {leverage: leverage})
				if (cmd.bc) {
					balance.leverage = leverage
					this.updateBalance(balance, currency, true, true)
				}
			}
		} catch(ex) {
			this.warn(`couldn't set leverage, please set manually! (Error: ${ex.error || ex.message})`)
		}

		order.externalReferralProgram = _('UHJvZml0Vmlldw')
		const result = yield* this.req('POST', order.triggerPrice || order.trailValue ? "/conditional_orders" : "/orders", order)
		cmd.id && !order.clientId && result && result.id && this.cacheId(result.id, cmd.id)
		return result
	}

	*ordersCancel(cmd) {
		const market = yield* this.symbolInfo(cmd.s)
		if (!cmd.ch && !cmd.d && !cmd.fp /*&& cmd.t != "open" && cmd.t != "close"*/ && !cmd.id && cmd.cm.wasPercent(true)) {
			let params = {market: market.name}
			if (cmd.b) params.side = cmd.isBid ? 'buy' : 'sell'
			if (cmd.tp || cmd.sl || cmd.so || cmd.ts || cmd.t === "close") {
				params.conditionalOrdersOnly = true
			} else if (cmd.t === "open") {
				params.limitOrdersOnly = true
			}
			this.info("canceling all "+params.market+(params.conditionalOrdersOnly ? " stop" : params.limitOrdersOnly ? " limit" : "")+" orders in bulk!")
			let success = true
			try {
				yield* this.req('DELETE', "/orders", params)
			} catch(ex) {
				success = false
				this.warn("bulk cancellation failed – trying one-by-one...")
			}
			if (success) return true
		}

		let orders = yield* this.req('GET', "/orders", {market: market.name})
		let stopOrders = yield* this.req('GET', "/conditional_orders", {market: market.name})

		if ((!orders || !orders.filter) && (!stopOrders || !stopOrders.filter)) {
			this.info("unable to get order info! (or no orders ever created!)")
			return false
		}
		orders = [...(orders||[]), ...(stopOrders||[])].filter(order => order.status !== "closed")

		if (!orders.length) {
			this.info("no orders open!")
			return false
		}

		const ids = (cmd.id || "").split(',')
		const hasType = cmd.has('t') && (cmd.t === "limit" || cmd.t === "market")
		const match = cmd.cr ? false : true
		orders = orders.filter(order => {
			if ((cmd.isBid && order.side !== 'buy') || (cmd.isAsk && order.side !== 'sell')) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(order.price)) {
				return !match
			}
			if ((cmd.tp && order.type !== "take_profit") || ((cmd.sl || cmd.so) && order.type !== "stop") || (cmd.ts && order.type !== "trailing_stop")) {
				return !match
			}
			if (hasType && !(order.orderType || order.type).includes(cmd.t)) {
				return !match
			}
			if (cmd.r && !order.reduceOnly) {
				return !match
			}
			if ((cmd.t === "open" && (Number(order.triggerPrice) || Number(order.trailValue))) || (cmd.t === "close" && !(Number(order.triggerPrice) || Number(order.trailValue)))) {
				return !match
			}
			if (cmd.id) {
				if (order.clientId !== undefined) {
					if (!order.clientId || !order.clientId.startsWithAny(ids, true)) {
						return !match
					}
				} else if (!this.checkId(order.id, ids)) {
					return !match
				}
			}
			return match
		})

		if (cmd.cm._().reference(orders.length).getMax() < orders.length) {
			const sort = {
				"newest": 	["createdAt", true],
				"oldest": 	["createdAt", false],
				"highest": 	["price", true],
				"lowest": 	["price", false],
				"biggest": 	["size", true],
				"smallest": ["size", false]
			}
			cmd.cmo === 'random' ? shuffle(orders) : sortByIndex(orders, sort[cmd.cmo][0], sort[cmd.cmo][1])
			orders = orders.slice(0, cmd.cm.resolve(0))
		}
		if(!orders.length) {
			this.info("open orders don't match any criteria!")
			return []
		}

		this.info((cmd.ch ? "orders found: " : "canceling orders: ")+stringify(orders))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return orders
		}

		let cancelOrders = []
		for (const order of orders) {
			const result = yield* this.req('DELETE', (order.triggerPrice || order.trailValue ? "/conditional_orders/" : "/orders/") + order.id)
			if (result === "Order queued for cancellation" || result === "Order cancelled") {
				cancelOrders.push(order)
				this.removeId(order.id)
			}
		}
		return cancelOrders
	}

	*positionsClose(cmd, pos, i, ohlc = {}, market) {
		const ticker = (cmd.pc && !cmd.up && this.getPrice(market.name)) || (yield* this.symbolTicker(market.name))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.name} is not available!`)
		}
		ticker.pos = pos.recentAverageOpenPrice || pos.entryPrice
		ticker.even = pos.recentBreakEvenPrice || pos.entryPrice
		ticker.liq = pos.estimatedLiquidationPrice || pos.entryPrice
		this.checkSlip(cmd, ohlc, market.name, ticker)
		cmd.up && this.updatePrice(market.name, ticker)

		const first = ticker[cmd.sl || cmd.tp || cmd.ts || cmd.so ? 'pos' : cmd.pr] || ohlc[cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(market.pricePrecision)
		if (cmd.fp) {
			price = cmd.fp.resolve(market.pricePrecision)
		}
		if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated, use minp= to set minimum)")
		}

		const leverage = cmd.l || pos.leverage || 1
		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		let order = {
			market: market.name,
			side: pos.side === 'buy' ? 'sell' : 'buy',
			price: cmd.t !== "market" ? price : null,
			type: cmd.t === "market" ? "market" : "limit",
			size: cmd.q._().mul(!cmd.q.wasPercent() ? leverage / (cmd.u === "currency" ? price : 1) : 1)
				.reference(pos.size).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.quantityPrecision),
			reduceOnly: cmd.r === undefined ? true : cmd.r
		}
		if (order.size <= 0) this.warn("your q parameter might be incorrect! (quantity of 0 calculated, use minq= to set minimum)")
		if (cmd.t === "ioc") order.ioc = true
		if (cmd.t === "post") order.postOnly = true

		if (cmd.tp) {
			cmd.tpref && cmd.tpref !== "price" && this.info("using custom TP reference: "+(ohlc[cmd.tpref] || ticker[cmd.tpref] || "NOT FOUND!")+" ("+cmd.tpref+")")
			cmd.tpref !== "price" && (price = (cmd.tpref && (ohlc[cmd.tpref] || ticker[cmd.tpref])) || first)
			order.type = "takeProfit"
			order.triggerPrice = cmd.tp._().relative(price).resolve(market.pricePrecision)
			order.price && (order.orderPrice = order.price)
			delete order.price
			delete order.ioc
			delete order.postOnly
			order.retryUntilFilled = cmd.rf === false ? false : true
		} else if (cmd.sl || cmd.so) {
			const ref = cmd.slref || cmd.soref
			ref && ref !== "price" && this.info("using custom Stop reference: "+(ohlc[ref] || ticker[ref] || "NOT FOUND!")+" ("+ref+")")
			ref !== "price" && (price = (ref && (ohlc[ref] || ticker[ref])) || first)
			order.type = "stop"
			order.triggerPrice = (cmd.sl || cmd.so)._().relative(price).resolve(market.pricePrecision)
			order.price && (order.orderPrice = order.price)
			delete order.price
			delete order.ioc
			delete order.postOnly
			order.retryUntilFilled = cmd.rf === false ? false : true
		} else if (cmd.ts) {
			cmd.tsref && cmd.tsref !== "price" && this.info("using custom TS reference: "+(ohlc[cmd.tsref] || ticker[cmd.tsref] || "NOT FOUND!")+" ("+cmd.tsref+")")
			cmd.tsref !== "price" && (price = (cmd.tsref && (ohlc[cmd.tsref] || ticker[cmd.tsref])) || first)
			order.type = "trailingStop"
			order.trailValue = cmd.ts._().reference(price).abs().mul(pos.side === 'buy' ? -1 : 1).resolve(market.pricePrecision)
			delete order.price
			delete order.ioc
			delete order.postOnly
			order.retryUntilFilled = cmd.rf === false ? false : true
		} else if (cmd.id) {
			order.clientId = this.uniqueID(cmd.id, i)
		}

		!cmd.ch && this.info(`placing Close Position order @ ${leverage}x: `+stringify(order))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}

		order.externalReferralProgram = _('UHJvZml0Vmlldw')
		const result = yield* this.req('POST', order.triggerPrice || order.trailValue ? "/conditional_orders" : "/orders", order)
		cmd.id && !order.clientId && result && result.id && this.cacheId(result.id, cmd.id)
		return result
	}

	*positionsCloseAll(cmd, ohlc, positions) {
		if (!positions) {
			positions = yield* this.req('GET', "/positions", {showAvgPrice: true})
			if (!positions || !positions.filter) {
				this.info("unable to get position info!")
				return false
			}
			const symbol = cmd.s && cmd.s !== '*' && this.sym(cmd.s)
			positions = positions.filter(pos => (!symbol || this.sym(pos.future) == symbol) && pos.size)
			if (!positions.length) {
				this.info("no positions open!")
				return false
			}
		}

		const symbols = yield* this.symbolInfo()
		const match = cmd.cr ? false : true
		positions = positions.filter(pos => {
			if ((cmd.isBid && pos.side !== 'buy') || (cmd.isAsk && pos.side !== 'sell')) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(pos.entryPrice)) {
				return !match
			}
			pos.leverage = Math.round(Math.abs(pos.cost) / pos.collateralUsed)
			if (cmd.l && pos.leverage !== cmd.l) {
				return !match
			}
			pos.symbol = this.sym(pos.future)
			const market = symbols[pos.symbol] || {}
			pos.currency = cmd.currency = (market.baseCurrency || this.meta.defaultSymbol) + marginSuffix
			pos._sizeUSD = pos.collateralUsed
			pos._sizeCoin = pos.size / pos.leverage
			pos.pnl = (pos.recentPnl || pos.recentPnl === 0 ? pos.recentPnl : pos.unrealizedPnl) / pos.collateralUsed / pos.leverage
			pos.avgEntryPrice = toPrecision(pos.recentAverageOpenPrice || pos.entryPrice || 0, market.pricePrecision)
			pos.markPrice = toPrecision(pos.avgEntryPrice * (1 + (pos.side === 'buy' ? pos.pnl : -pos.pnl)), market.pricePrecision)
			if (!this.checkPL(cmd, ohlc, pos, balance => balance / pos.avgEntryPrice * pos.leverage, market.quantityPrecision)) {
				return !match
			}
			return match
		})

		if (cmd.cm._().reference(positions.length).getMax() < positions.length) {
			if (["newest", "oldest"].includes(cmd.cmo)) {
				this.warn(`Close Maximum Order option "${cmd.cmo}" is currently not supported!`)
			} else {
				const sort = {
					"highest": 	["entryPrice", true],
					"lowest": 	["entryPrice", false],
					"biggest": 	["size", true],
					"smallest": ["size", false]
				}
				cmd.cmo === 'random' ? shuffle(positions) : sortByIndex(positions, sort[cmd.cmo][0], sort[cmd.cmo][1])
				positions = positions.slice(0, cmd.cm.resolve(0))
			}
		}
		if(!positions.length) {
			this.info("open positions don't match any criteria!")
			return []
		} else if (cmd.ch) {
			this.info(`${positions.length} position${positions.length > 1 ? 's' : ''} found!`)
			return positions
		}

		let closeOrders = [], i = 0
		for (const pos of positions) {
			const order = yield* this.positionsClose(cmd, pos, i++, ohlc, symbols[pos.symbol])
			if (order) {
				closeOrders.push(order)
				Object.assign(closeOrders.last(), Object.without(pos, ["size", "side"]))
			}
		}
		return closeOrders
	}

	*transfer(cmd, ohlc) {
		const response = yield* this.req('GET', "/subaccounts")
		if (!response || !response.filter) {
			this.info("unable to get sub account information!")
			return false
		}
		if (!response.length) {
			this.info("no sub accounts found!")
			return false
		}
		let subs = {}
		response.forEach(info => subs[info.nickname.toLowerCase()] = info)

		const y = cmd.y.split(':')
		let src = y.length > 1 ? y[0] : 'main'
		let dest = y.length > 1 ? y[1] : y[0]

		if (!subs[src] && src !== 'main') {
			throw new ReferenceError(`Sub account with nickname "${src}" not found! (available: ${Object.mapAsArray(subs, (sub, info) => info.nickname).join(', ')})`)
		} else if (!subs[dest] && dest !== 'main') {
			throw new ReferenceError(`Sub account with nickname "${dest}" not found! (available: ${Object.mapAsArray(subs, (sub, info) => info.nickname).join(', ')})`)
		}
		src = src !== 'main' ? subs[src].nickname : src
		dest = dest !== 'main' ? subs[dest].nickname : dest

		let balances = {}
		if (cmd.q.wasPercent()) {
			if (src === 'main') {
				balances = yield* this.account(cmd.tb[0])
				this.updateBalance(balances, cmd.tb[0])
			} else {
				const response = yield* this.req('GET', `/subaccounts/${src}/balances`)
				if (!response || !response.filter || !response.length) {
					this.info(`unable to get sub account balances for "${src}"!`)
					return false
				}
				response.forEach(info => balances[info.coin] = info)
			}
		}

		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		let orders = []
		for (const currency of cmd.tb) {
			if (cmd.q.wasPercent() && !balances[currency]) {
				throw new ReferenceError(`Account Balance ${currency} not available in "${src}"!`)
			}
			let balance = balances[currency] || {}
			balance = balance.free || balance.available || 0
			const order = {
				coin: currency,
				size: cmd.q._().reference(balance).mul(ohlc[cmd.qr] || 1).resolve(8),
				source: src,
				destination: dest
			}
			this.info(`transferring ${order.size}${balance ? ` of ${balance}` : ''} ${currency} from "${src}" to "${dest}"!`)

			if (cmd.d) {
				this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
				orders.push(order)
			} else {
				orders.push(yield* this.req('POST', "/subaccounts/transfer", order))
			}
		}
		return orders
	}

	*convert(cmd) {
		const balances = cmd.q.wasPercent() ? (yield* this.account()) : {}
		let orders = []

		for (let fromTo of cmd.convert) {
			fromTo = fromTo.split(':')
			const currency = fromTo[0]
			if (cmd.q.wasPercent() && !balances[currency]) {
				throw new ReferenceError(`Account Balance ${currency} not available!`)
			}
			const balance = balances[currency] || {}
			const order = {
				fromCoin: currency,
				toCoin: fromTo[1],
				size: cmd.q._().reference(balance.free || balance.available || 0).resolve(8)
			}
			if (order.size <= 0) {
				this.warn(`conversion amount for ${currency} is zero or less: ${order.size}!`)
				continue
			}
			this.info(`converting ${order.size} ${currency} to ${order.toCoin}!`)

			if (cmd.d) {
				this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
				orders.push(order)
			} else {
				const quote = yield* this.req('POST', "/otc/quotes", order)
				let status = null
				do {
					while((yield* sleep(0.5)) == null);
					status = yield* this.req('GET', "/otc/quotes/"+quote.quoteId)
				} while(status && !status.cost && !status.expired)
				if (!status || status.expired) {
					this.warn('conversion quote expired!')
				} else {
					orders.push(Object.assign(order, status))
					const result = yield* this.req('POST', `/otc/quotes/${quote.quoteId}/accept`)
				}
			}
		}
		return orders
	}
}

Broker.add(new FTX())
