"use strict"

class OANDASandbox extends Exchange {
	static meta = {
		aliases: [
			"OANDASANDBOX",
			"OANDA-SANDBOX",
		],
		endpoint: "https://api-fxpractice.oanda.com",
		fields: {
			public: {
				label: "Account Number",
				type: "text",
			},
			private: {
				label: "Auth Token",
			},
		},
		name: "OANDA Sandbox",
		patterns: [],
		permissions: {
			origins: [
				"https://api-fxpractice.oanda.com/*",
			],
			permissions: []
		},
		subscriptions: {
			active: [],
			inactive: [],
		},
		version: "v3",
		website: "https://www.oanda.com/",
		desc: "Award-winning leader in Currency Data, Forex & CFD Trading, offering leveraged trading, payment and data services for businesses and investors.",
		defaultSymbol: "USD",
		testSymbol: "USD_CHF",
		symbolAuth: true, margin: true, stopsOnFill: true, reducePosOnly: true,
	}
	constructor() {
		super(OANDASandbox.meta)
	}

	*req(method, resource = "", params) {
		const credentials = this.getCredentials("private")
		const headers = {
			"Accept-Datetime-Format": "UNIX",
			"Authorization": "Bearer " + credentials.private,
			"Content-Type": "application/json"
		}
		resource = "/" + this.meta.version + "/accounts/" + credentials.public + resource
		if (params && method === 'GET') {
			resource += '?'+serialize(params).replace(/%20/g, '+')
			params = null
		}
		params = params ? JSON.stringify(params) : ''

		const response = yield request.bind(this, method, this.meta.endpoint + resource, params, headers)
		if (!response) {
			throw new Error("Unknown connection error with API endpoint!")
		}
		if (response.errorMessage) {
			throw new Error(response.errorMessage + (response.errorCode ? ` (${response.errorCode})` : ''))
		}
		return response
	}

	*account() {
		const account = (yield* this.req('GET')).account || {}

		return {
			available: Number(account.marginAvailable || 0),
			balance: Number(account.balance || 0),
			currency: account.currency,
			leverage: 1 / Number(account.marginRate)
		}
	}

	*symbolInfo(symbol) {
		const symbolCache = this.getSymbolCache()
		if (symbolCache._refresh) {
			const response = yield* this.req('GET', "/instruments")

			response.instruments.forEach(info => {
				info.leverage = Math.round(1 / Number(info.marginRate || 0.01))
				symbolCache[info.name.replace('_', '')] = info
				symbolCache[info.displayName.toUpperCase().replace('/', '')] = info
			})
			this.updatedSymbolCache()
		}
		if (!symbol) return symbolCache

		const lookup = this.sym(symbol)
		if (!symbolCache[lookup]) {
			throw new Error(`Unknown market symbol: ${symbol} – Use the 'Symbols' button in the exchange settings or alert editor to browse for valid symbols!`)
		}
		return symbolCache[lookup]
	}

	*symbolTicker(symbol) {
		let ticker = yield* this.req('GET', "/pricing", {instruments: symbol})
		ticker = ticker && ticker.prices && ticker.prices[0]

		return ticker && {
			ask: Number(ticker.asks[0].price),
			bid: Number(ticker.bids[0].price),
			mid: (Number(ticker.asks[0].price) + Number(ticker.bids[0].price)) / 2
		}
	}

	*convert(from, to, useCache) {
		const symbols = this.getSymbolCache()
		let inv = false
		const direct = symbols[from + to] || ((inv = true) && symbols[to + from])
		const directTicker = direct && ((useCache && this.getPrice(direct.name)) || (yield* this.symbolTicker(direct.name)))
		if (directTicker) {
			return inv ? 1 / directTicker.mid : directTicker.mid
		}

		inv = false
		const toUSD = symbols[from + 'USD'] || ((inv = true) && symbols['USD' + from])
		const toUSDTicker = toUSD && ((useCache && this.getPrice(toUSD.name)) || (yield* this.symbolTicker(toUSD.name)))
		if (!toUSDTicker) {
			return false
		}
		let conv = inv ? 1 / toUSDTicker.mid : toUSDTicker.mid

		inv = false
		const fromUSD = symbols[to + 'USD'] || ((inv = true) && symbols['USD' + to])
		const fromUSDTicker = fromUSD && ((useCache && this.getPrice(fromUSD.name)) || (yield* this.symbolTicker(fromUSD.name)))
		if (!fromUSDTicker) {
			return false
		}
		conv *= inv ? fromUSDTicker.mid : 1 / fromUSDTicker.mid
		return conv
	}

	*trade(cmd, ohlc = {}) {
		const market = yield* this.symbolInfo(cmd.s)
		cmd._sym = market.name
		const ticker = (cmd.pc && !cmd.up && this.getPrice(market.name)) || (yield* this.symbolTicker(market.name))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.name} is not available!`)
		}
		this.checkSlip(cmd, ohlc, market.name, ticker)
		if (cmd.up) {
			this.updatePrice(market.name, ticker)
			if (!cmd.b && !cmd.ub) {
				return ticker
			}
		}

		let balance = {available: 0, balance: 0}
		if ((cmd.y !== "possize" && cmd.q.wasPercent()) || cmd.ub) {
			balance = cmd.bc && !cmd.ub && this.getBalance() || (yield* this.account())
			this.checkBalance(cmd, balance, balance.currency)
			if (!cmd.bc || cmd.ub) {
				this.updateBalance(balance, balance.currency)
				if (cmd.ub) {
					return balance
				}
			}
		}

		const precision = market.displayPrecision || 5
		const first = ticker[cmd.pr] || ohlc[cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.fp ? cmd.fp.resolve(precision) : cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(precision)
		if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated, use minp= to set minimum)")
		}

		ohlc[cmd.lr] && this.info(`using custom leverage multiplier: ${ohlc[cmd.lr]} (${cmd.lr})`)
		let cf = 1.0
		let leverage = cmd._lev = cmd.lc || cmd.l * (ohlc[cmd.lr] || 1) || balance.leverage || (this.getBalance("", false, true) || {}).leverage || 1
		if (leverage > market.leverage) {
			cmd.l && (cf = market.leverage / leverage)
			leverage = cmd._lev = market.leverage
			this.warn(`leverage limited to ${leverage}x by instrument!`)
		}
		const available = cmd.y === "equity" ? balance.balance : balance.available
		let quantity = ohlc[cmd.y] || available / price * leverage * cf
		const baseCurrency = cmd.s.slice(-3)
		if (cmd.y !== "possize" && cmd.q.wasPercent() && balance.currency !== baseCurrency) {
			const conv = yield* this.convert(balance.currency, baseCurrency, cmd.pc && !cmd.up)
			if (conv && conv != 1) {
				this.info(`converting account currency ${balance.currency} to symbol base currency ${baseCurrency}: ${toPrecision(conv, 6)} (factor)`)
				quantity *= conv
				this.setBalanceRaw(balance.currency, setObject(baseCurrency, conv))
			} else {
				this.warn(`could not convert account currency ${balance.currency} to symbol base currency ${baseCurrency}, please use absolute units for buy in!`)
			}
		}

		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		quantity = cmd.q._().mul(cmd.u === "currency" && !cmd.q.wasPercent() ? leverage * cf / price : 1)
			.reference(quantity).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.tradeUnitsPrecision || 0)
		if (cmd.lq && ohlc.left) {
			this.info("using leftover quantity from last order: "+ohlc.left)
			quantity = ohlc.left.toFixed(market.tradeUnitsPrecision || 0)
		}
		if (market.minimumTradeSize && quantity < market.minimumTradeSize) {
			this.warn(`order quantity below instrument minimum of ${market.minimumTradeSize} – use minq= to set minimum!`)
		} else if (quantity <= 0) {
			this.warn("your q parameter might be incorrect! (quantity of 0 calculated, use minq= to set minimum)")
		}

		let order = {
			instrument: market.name,
			positionFill: cmd.r ? "REDUCE_ONLY" : "DEFAULT",
			triggerCondition: ({mid: "MID", bid: "BID", ask: "ASK"})[cmd.pr] || "DEFAULT",
			units: `${cmd.isAsk ? '-' : ''}${quantity}`,
			type: cmd.t === "market" ? "MARKET" : "LIMIT",
			timeInForce: ({fok: "FOK", ioc: "IOC", day: "GFD", market: "FOK"})[cmd.t] || "GTC"
		}
		if (cmd.t !== "market") order.price = price
		if (cmd.pb) order.priceBound = cmd.pb._().relative(first).resolve(precision)
/*		if (cmd.gtd) {
			order.timeInForce = "GTD"
			order.gtdTime = cmd.gtd
		}*/

		if (cmd.so) {
			cmd.soref && cmd.soref !== "price" && this.info(`using custom Stop reference: ${ohlc[cmd.soref] || ticker[cmd.soref] || "NOT FOUND!"} (${cmd.soref})`)
			const stop = cmd.soref === "price" ? price : ((cmd.soref && (ohlc[cmd.soref] || ticker[cmd.soref])) || first)
			order.price = cmd.so._().relative(stop).resolve(precision)
			order.type = "STOP"
		}
		if (cmd.sl) {
			cmd.slref && cmd.slref !== "price" && this.info(`using custom SL reference: ${ohlc[cmd.slref] || ticker[cmd.slref] || "NOT FOUND!"} (${cmd.slref})`)
			const stop = cmd.slref === "price" ? price : ((cmd.slref && (ohlc[cmd.slref] || ticker[cmd.slref])) || first)
			order[cmd.gsl ? 'guaranteedStopLossOnFill' : 'stopLossOnFill'] = {price: cmd.sl._().relative(stop).resolve(precision)}
			if (cmd.id) {
				order[cmd.gsl ? 'guaranteedStopLossOnFill' : 'stopLossOnFill'].clientExtensions = {id: this.uniqueID(cmd.id + 'SL')}
			}
		}
		if (cmd.tp) {
			cmd.tpref && cmd.tpref !== "price" && this.info(`using custom TP reference: ${ohlc[cmd.tpref] || ticker[cmd.tpref] || "NOT FOUND!"} (${cmd.tpref})`)
			const stop = cmd.tpref === "price" ? price : ((cmd.tpref && (ohlc[cmd.tpref] || ticker[cmd.tpref])) || first)
			order.takeProfitOnFill = {price: cmd.tp._().relative(stop).resolve(precision)}
			if (cmd.id) {
				order.takeProfitOnFill.clientExtensions = {id: this.uniqueID(cmd.id + 'TP')}
			}
		}
		if (cmd.ts) {
			cmd.tsref && cmd.tsref !== "price" && this.info(`using custom TS reference: ${ohlc[cmd.tsref] || ticker[cmd.tsref] || "NOT FOUND!"} (${cmd.tsref})`)
			const stop = cmd.tsref === "price" ? price : ((cmd.tsref && (ohlc[cmd.tsref] || ticker[cmd.tsref])) || first)
			order.trailingStopLossOnFill = {distance: cmd.ts._().reference(stop).abs().resolve(precision)}
			if (cmd.id) {
				order.trailingStopLossOnFill.clientExtensions = {id: this.uniqueID(cmd.id + 'TS')}
			}
		}
		if (cmd.id) {
			order.clientExtensions = {id: this.uniqueID(cmd.id)}
		}

		this.info(`placing order @ ${leverage}x: `+stringify(order))
		if (cmd.d) {
			this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}
		try {
			if (cmd.l !== undefined) {
				this.info(`setting account leverage to ${leverage}x...`)
				yield* this.req('PATCH', "/configuration", {marginRate: (1 / leverage).toFixed(2)})
			}
		} catch(ex) {
			this.warn(`couldn't set leverage, please set manually! (Error: ${ex.message})`)
		}
		return yield* this.ordersPlace(order)
	}

	*ordersPlace(order) {
		const result = yield* this.req('POST', "/orders", {order: order})

		if (result.orderCancelTransaction) {
			throw new Error(result.orderCancelTransaction.reason || "Unknown reason")
		}
		return result.orderFillTransaction || result.orderCreateTransaction || result
	}

	*ordersCancel(cmd) {
		const market = yield* this.symbolInfo(cmd.s)
		cmd._sym = market.name
		let orders = ((yield* this.req('GET', "/orders", {
			state: "PENDING",
			instrument: market.name,
			count: 500
		})) || {}).orders

		if (!orders || !orders.filter) {
			this.info("unable to get order info!")
			return false
		} else if (!orders.length) {
			this.info("no orders open!")
			return false
		}

		const ids = (cmd.id || "").split(',')
		const match = cmd.cr ? false : true
		orders = orders.filter(order => {
			if ((cmd.isBid && order.units < 0) || (cmd.isAsk && order.units > 0)) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(order.price)) {
				return !match
			}
			if ((cmd.t === "open" && order.type !== "LIMIT") || (cmd.t === "close" && order.type === "LIMIT")) {
				return !match
			}
			if ((cmd.tp && order.type !== "TAKE_PROFIT") || (cmd.sl && order.type !== "STOP_LOSS") || (cmd.so && order.type !== "STOP") || (cmd.ts && order.type !== "TRAILING_STOP_LOSS")) {
				return !match
			}
			if (cmd.r !== undefined && !cmd.r != (order.positionFill != "REDUCE_ONLY")) {
				return !match
			}
			if (cmd.id && (!order.clientExtensions || !order.clientExtensions.id || !order.clientExtensions.id.startsWithAny(ids, true))) {
				return !match
			}
			return match
		})

		if (cmd.cm._().reference(orders.length).getMax() < orders.length) {
			const sort = {
				"newest": 	["createTime", true],
				"oldest": 	["createTime", false],
				"highest": 	["price", true],
				"lowest": 	["price", false],
				"biggest": 	["units", true],
				"smallest": ["units", false]
			}
			cmd.cmo === 'random' ? shuffle(orders) : sortByIndex(orders, sort[cmd.cmo][0], sort[cmd.cmo][1])
			orders = orders.slice(0, cmd.cm.resolve(0))
		}
		if(!orders.length) {
			this.info("open orders don't match any criteria!")
			return []
		}

		this.info((cmd.ch ? "orders found: " : "canceling orders: ")+stringify(orders))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return orders
		}

		let cancelOrders = []
		for (const order of orders) {
			const result = yield* this.req('PUT', `/orders/${order.id}/cancel`)
			if (result && result.orderCancelTransaction && result.orderCancelTransaction.orderID === order.id) {
				cancelOrders.push(order)
			}
		}
		return cancelOrders
	}

	*positionsClose(cmd, pos, i, ohlc = {}, market) {
		const ticker = (cmd.pc && !cmd.up && this.getPrice(market.name)) || (yield* this.symbolTicker(market.name))
		if (!ticker) {
			throw new ReferenceError(`Ticker ${market.name} is not available!`)
		}
		const precision = market.displayPrecision || 5
		ticker.pos = pos.averagePrice
		this.checkSlip(cmd, ohlc, market.name, ticker)
		cmd.up && this.updatePrice(market.name, ticker)

		const first = ticker[cmd.sl || cmd.tp || cmd.ts ? 'pos' : cmd.pr] || ohlc[cmd.pr] || ((cmd.isBid && cmd.t !== "market") || (cmd.isAsk && cmd.t === "market") ? ticker.bid : ticker.ask)
		let price = cmd.p._().relative(first).minmax(cmd.minp, cmd.maxp).resolve(precision)
		if (cmd.fp) {
			price = cmd.fp.resolve(precision)
		}
		if (price <= 0) {
			this.warn("your p parameter might be incorrect! (price of 0 calculated, use minp= to set minimum)")
		}

		const leverage = cmd.l || pos.leverage || 1
		ohlc[cmd.qr] && this.info(`using custom quantity multiplier: ${ohlc[cmd.qr]} (${cmd.qr})`)
		const quantity = cmd.q._().mul(cmd.u === "currency" && !cmd.q.wasPercent() ? leverage / price : 1)
			.reference(pos.quantity).mul(ohlc[cmd.qr] || 1).minmax(cmd.minq, cmd.maxq).resolve(market.tradeUnitsPrecision || 0)
		if (market.minimumTradeSize && quantity < market.minimumTradeSize) {
			this.warn(`order quantity below instrument minimum of ${market.minimumTradeSize} – use minq= to set minimum!`)
		} else if (quantity <= 0) {
			this.warn("your q parameter might be incorrect! (quantity of 0 calculated, use minq= to set minimum)")
		}

		if (cmd.so || cmd.sl || cmd.tp || cmd.ts || (cmd.t === "market" && cmd.r !== false && quantity >= pos.quantity)) {
			let closeOrders = []

			// $$$$ make "selectable" which tradeIDs!
			for (const tradeID of pos.tradeIDs) {
				let order = {tradeID: tradeID}

				if (cmd.so) {
					cmd.soref && cmd.soref !== "price" && this.info(`using custom Stop reference: ${ohlc[cmd.soref] || ticker[cmd.soref] || "NOT FOUND!"} (${cmd.soref})`)
					const stop = cmd.soref === "price" ? price : ((cmd.soref && (ohlc[cmd.soref] || ticker[cmd.soref])) || first)
					order.type = "STOP"
					order.price = cmd.so._().relative(stop).resolve(precision)
				} else if (cmd.sl) {
					cmd.slref && cmd.slref !== "price" && this.info(`using custom SL reference: ${ohlc[cmd.slref] || ticker[cmd.slref] || "NOT FOUND!"} (${cmd.slref})`)
					const stop = cmd.slref === "price" ? price : ((cmd.slref && (ohlc[cmd.slref] || ticker[cmd.slref])) || first)
					order.type = cmd.gsl ? "GUARANTEED_STOP_LOSS" : "STOP_LOSS"
					order.price = cmd.sl._().relative(stop).resolve(precision)
				} else if (cmd.tp) {
					cmd.tpref && cmd.tpref !== "price" && this.info(`using custom TP reference: ${ohlc[cmd.tpref] || ticker[cmd.tpref] || "NOT FOUND!"} (${cmd.tpref})`)
					const stop = cmd.tpref === "price" ? price : ((cmd.tpref && (ohlc[cmd.tpref] || ticker[cmd.tpref])) || first)
					order.type = "TAKE_PROFIT"
					order.price = cmd.tp._().relative(stop).resolve(precision)
				} else if (cmd.ts) {
					cmd.tsref && cmd.tsref !== "price" && this.info(`using custom TS reference: ${ohlc[cmd.tsref] || ticker[cmd.tsref] || "NOT FOUND!"} (${cmd.tsref})`)
					const stop = cmd.tsref === "price" ? price : ((cmd.tsref && (ohlc[cmd.tsref] || ticker[cmd.tsref])) || first)
					order.type = "TRAILING_STOP_LOSS"
					order.distance = cmd.ts._().reference(stop).abs().resolve(precision)
				} else {
					order.units = "ALL"
				}
				if (cmd.pb) order.priceBound = cmd.pb._().relative(first).resolve(precision)
/*				if (cmd.gtd) {
					order.timeInForce = "GTD"
					order.gtdTime = cmd.gtd
				}*/
				if (cmd.id) {
					order.clientExtensions = {id: this.uniqueID(cmd.id, i, closeOrders.length)}
				}
				closeOrders.push(order)
			}

			!cmd.ch && this.info(`placing Close Position order @ ${leverage}x: `+stringify(closeOrders))
			if (cmd.ch || cmd.d) {
				!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
				return closeOrders
			}

			let results = []
			for (const order of closeOrders) {
				if (order.units) {
					const result = yield* this.req('PUT', `/trades/${order.tradeID}/close`, order)
					results.push(result.orderFillTransaction || result.orderCreateTransaction || result)
				} else {
					results.push(yield* this.ordersPlace(order))
				}
			}
			return results
		}

		let order = {
			instrument: market.name,
			positionFill: cmd.r !== false ? "REDUCE_ONLY" : "REDUCE_FIRST",
			triggerCondition: ({mid: "MID", bid: "BID", ask: "ASK"})[cmd.pr] || "DEFAULT",
			units: `${pos.units > 0 ? '-' : ''}${quantity}`,
			type: cmd.t === "market" ? "MARKET" : "LIMIT",
			timeInForce: ({fok: "FOK", ioc: "IOC", day: "GFD", market: "FOK"})[cmd.t] || "GTC"
		}
		if (cmd.t !== "market") order.price = price
		if (cmd.pb) order.priceBound = cmd.pb._().relative(first).resolve(precision)
/*		if (cmd.gtd) {
			order.timeInForce = "GTD"
			order.gtdTime = cmd.gtd
		}*/
		if (cmd.id) {
			order.clientExtensions = {id: this.uniqueID(cmd.id, i)}
		}

		!cmd.ch && this.info(`placing Close Position order @ ${leverage}x: `+stringify(order))
		if (cmd.ch || cmd.d) {
			!cmd.ch && this.warn("nothing done! (d=1 [disabled / test mode] was specified)")
			return order
		}
		return yield* this.ordersPlace(order)
	}

	*positionsCloseAll(cmd, ohlc, positions) {
		if (!positions) {
			positions = yield* this.req('GET', "/openPositions")
			positions = positions && positions.positions
			if (!positions || !positions.filter) {
				this.info("unable to get position info!")
				return false
			}
			const symbol = cmd.s && cmd.s !== '*' && (yield* this.symbolInfo(cmd.s)).name
			positions = positions.filter(pos => {
				Object.assign(pos, pos.long.units > 0 ? pos.long : pos.short)
				delete pos.long
				delete pos.short
				return (!symbol || pos.instrument === symbol) && (pos.units = Number(pos.units)) !== 0
			})
			if (!positions.length) {
				this.info("no positions open!")
				return false
			}
		}

		const symbols = yield* this.symbolInfo()
		const match = cmd.cr ? false : true
		positions = positions.filter(pos => {
			if ((cmd.isBid && pos.units < 0) || (cmd.isAsk && pos.units > 0)) {
				return false
			}
			if (cmd.fp && cmd.fp.compare(pos.averagePrice)) {
				return !match
			}
			pos.symbol = this.sym(pos.instrument)
			const market = symbols[pos.symbol] || {}
			cmd._sym = market.name
			// $$$$ check pos monitor => balance.lastSymbol ???
			pos.currency = cmd.currency = ""
			const balance = this.getBalance(cmd.currency, false, true) || {}
			const baseCurrency = pos.symbol.slice(-3)
			const conv = balance[baseCurrency] || 1
			pos.side = pos.units > 0 ? 'long' : 'short'
			pos.quantity = Math.abs(pos.units)
			pos._sizeUSD = Number(pos.marginUsed)
			pos.leverage = (pos.quantity * pos.averagePrice / conv) / pos._sizeUSD
			pos.pnl = pos.unrealizedPL / pos._sizeUSD / pos.leverage
			pos.markPrice = pos.markPrice || toPrecision(pos.averagePrice * (1 + pos.pnl), market.tradeUnitsPrecision)
			pos.leverage = Math.round(pos.leverage)
			if (!this.checkPL(cmd, ohlc, pos, total => total / pos.averagePrice * conv * pos.leverage, market.tradeUnitsPrecision)) {
				return !match
			}
			return match
		})

		if (cmd.cm._().reference(positions.length).getMax() < positions.length) {
			if (["newest", "oldest"].includes(cmd.cmo)) {
				this.warn(`Close Maximum Order option "${cmd.cmo}" is currently not supported!`)
			} else {
				const sort = {
					"highest": 	["averagePrice", true],
					"lowest": 	["averagePrice", false],
					"biggest": 	["quantity", true],
					"smallest": ["quantity", false]
				}
				cmd.cmo === 'random' ? shuffle(positions) : sortByIndex(positions, sort[cmd.cmo][0], sort[cmd.cmo][1])
				positions = positions.slice(0, cmd.cm.resolve(0))
			}
		}
		if(!positions.length) {
			this.info("open positions don't match any criteria!")
			return []
		} else if (cmd.ch) {
			this.info(`${positions.length} position${positions.length > 1 ? 's' : ''} found!`)
			return positions
		}

		let closeOrders = [], i = 0
		for (const pos of positions) {
			const order = yield* this.positionsClose(cmd, pos, i++, ohlc, symbols[pos.symbol])
			if (order) {
				Array.isArray(order) ? closeOrders.push(...order) : closeOrders.push(order)
				Object.assign(closeOrders.last(), Object.without(pos, ["units", "side"]))
			}
		}
		return closeOrders
	}
}

Broker.add(new OANDASandbox())
